#pragma once
#include <ts/coverage.hpp>
#include <utils/utils.hpp>
#include <c_tracer/c_tracer.h>

#ifdef CPPTESK_BUILD_WITH_OLD_ASPECTRACE_PATH
extern "C" {
#include <aspectrace.h>
#include <aspect_coverage.h>
}
#else
extern "C" {
#include <aspectrace/aspectrace.h>
#include <aspectrace/aspect_coverage.h>
}
#endif //CPPTESK_BUILD_WITH_OLD_ASPECTRACE_PATH

namespace cpptesk
{
namespace tracer
{

typedef cpptesk::utils::ExecuteOnBlockEnd<void(*)(void)> TracerEndFunction;

namespace _internal
{
void dummy();
}

namespace coverage
{
using namespace cpptesk::ts::coverage;

template <typename C>
void aspectraceCoverageStructure()
{
    // The code below let us know the coverage size
    int size = 0;
    for(typename C::ElementIterator i = C::begin(); i != C::end(); ++i) size++;

    // To save aspect_coverage_elements + NULL terminator
    aspect_coverage_element **covelements = (aspect_coverage_element**)malloc(sizeof(char*) * (size + 1) * 3);

    // Extracting coverage from CPPTESK coverage to aspectrace coverage
    int j = 0;
    for(typename C::ElementIterator i = C::begin(); i != C::end(); ++i) {
        covelements[j++] = create_CoverageElement((*i).label.c_str(), NULL);
    }
    covelements[j] = NULL;

    // Print everything to file
    std::string name = C::name();
    aspect_coverage_traceCoverageStructure_a("coverage", name.c_str(), covelements);

    free(covelements);
}

/// @brief Trace structure of the specified coverage.
///
/// @tparam Type of the coverage to trace.
template <typename C>
void traceCoverageStructure()
{
    static int lastTraceId = -1;
    if (lastTraceId >= ::getTraceId()) return;
    lastTraceId = ::getTraceId();

    ::traceCoverageStart(C::name());
    TracerEndFunction _coverageEnd(::traceCoverageEnd);
    int i = 0;
    for (typename C::ElementIterator iter = C::begin(); iter != C::end(); ++iter)
    {
        ::traceElement(i++, (*iter).label.c_str());
    }
}

template<typename C>
struct CoverageStructure
{
    typedef typename C::Element Element;
    typedef typename C::ElementIterator ElementIterator;

    const char * name()
    {
        return C::name();
    }
    ElementIterator begin()
    {
        return C::begin();
    }
    ElementIterator end()
    {
        return C::end();
    }
    void trace()
    {
        static int lastTraceId = -1;
        if (lastTraceId >= ::getTraceId()) return;
        lastTraceId = ::getTraceId();

        ::traceCoverageStart(C::name());
        TracerEndFunction _coverageEnd(::traceCoverageEnd);
        int i = 0;
        for (typename C::ElementIterator iter = C::begin(); iter != C::end(); ++iter)
        {
            ::traceElement(i, (*iter).label);

        }

    }
};


///  
template<typename C>
struct TraceableCounter
{
    typedef typename C::Id Id;
private:
    Id _id;
    unsigned int _count;
    bool _must_trace;
    bool _is_set;
    bool _was_dumped_to_aspectrace;
    void trace()
    {
        if (_must_trace)
        {
            typename C::ElementIterator i = findElement<C>(_id, false);
            if (i != C::end())
            {
                // TO BE RESTORED AFTER UNDERSTANDING
                // ::traceCoverageElement("", C::name(), std::distance(C::begin(), i));

                // Aspectrace tracing
                if(!_was_dumped_to_aspectrace)
                {
                    std::string name = C::name();
                    aspect_coverage_traceCoverageElement("coverage", name.c_str(), (*i).label.c_str(), NULL);
                    _was_dumped_to_aspectrace = true;
                }
            }
        }
    }
public:
    operator unsigned int() const { return _count; }

    TraceableCounter& operator ++()
    {
        trace();
        ++_count;
        return *this;
    }

    unsigned int operator ++(int)
    {
        unsigned int result = _count;
        operator++();
        return result;
    }

    TraceableCounter() : _id(), _count(0), _must_trace(false), _is_set(false), _was_dumped_to_aspectrace(false) {}

    inline bool is_set() const { return _is_set; }
    inline bool must_trace() const { return _must_trace; }
    inline TraceableCounter& must_trace(bool must) { _must_trace = must; return *this; }
    inline TraceableCounter& set(const Id& id) { _id = id; _is_set = true; return *this; }

};

template<typename C>
struct SingleCoverageTracker
{
    typedef typename C::Id Id;
    typedef TraceableCounter<C> count_t;
    typedef std::map<Id, count_t> Counter;

protected:
    Counter _counter;
    bool	_must_trace;

public:
    SingleCoverageTracker(bool must_trace = true) : _counter(), _must_trace(must_trace)  { if(_must_trace) aspectraceCoverageStructure<C>(); }
private:
    SingleCoverageTracker(const SingleCoverageTracker& other) : _counter(other._counter) {}
    SingleCoverageTracker& operator= (const SingleCoverageTracker& other) ;
public:

    count_t& operator[](const Id& id)
    {
        count_t& result = _counter[id];
        if (!result.is_set()) result.set(id).must_trace(_must_trace);
        return result;
    }

    SingleCoverageTracker& operator << (const Id& id)
    {
        ++(this->operator[](id)) ;
        return *this;
    }

    SingleCoverageTracker& operator << (const C& cov)
    {
        *this << cov.id;
        return *this;
    }
    count_t& operator[](const C& cov)
    {
        return this->operator[](cov.id);
    }
};

struct CoverageTracker
{
    typedef unsigned int count_t;
private:
    CoverageTracker(const CoverageTracker& other) {}
protected:
    bool _must_trace;
    template<typename T>
    SingleCoverageTracker<T>& _getTracker()
    {
        static SingleCoverageTracker<T> tracker(_must_trace);
        return tracker;
    }

public:
    CoverageTracker(bool must_trace = true) : _must_trace(must_trace) {}
    template<typename T>
    SingleCoverageTracker<T>& tracker()
    {
        return _getTracker<T>();
    }
    template<typename T>
    CoverageTracker& operator<< (const T& id)
    {
        _getTracker<T>() << id;
        return *this;
    }
    template<typename T>
    typename SingleCoverageTracker<T>::count_t& operator[](const T& id)
    {
        return tracker<T>()[id];
    }
};

}
}
}
