/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of an operation execution process
///////////////////////////////////////////////////////////////////////////////

#include <hw/process.hpp>

#include <hw/assertion.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ucontext.h>

#ifdef CPPTESK_CLONE_INPUT_MESSAGE
    #define _CPPTESK_CLONE_MESSAGE(message)  ((message)->clone())
    #define _CPPTESK_DELETE_MESSAGE(message) if(message) { delete (message); }
#else
    #define _CPPTESK_CLONE_MESSAGE(message)  (message)
    #define _CPPTESK_DELETE_MESSAGE(message)
#endif

#define _CPPTESK_PROCESS_STACK_SIZE (32*1024)

namespace cpptesk {
namespace hw {

Process::Process(pid_t pid, Interface &iface, Message &message):
    pid(pid),
    priority(LOWEST_PRIORITY),
    status(EXECUTING),
    syn_point(START),
    clk_start(0),
    clk_total(0),
    clk_count(0),
    timeout(false),
    input_iface(&iface),
    input_message(&message),
    original_message(&message)
{
    this->line = -1;
    this->file = NULL;

    input_message = message.clone();
    CPPTESK_ASSERTION(input_message != 0, "Input message has not been cloned");

    ucp = (ucontext_t*)malloc(sizeof(ucontext_t));
    CPPTESK_ASSERTION(ucp != NULL, "No memory");

    getcontext(ucp);
    ucp->uc_stack.ss_size = _CPPTESK_PROCESS_STACK_SIZE;
    ucp->uc_stack.ss_sp = (char *)malloc(_CPPTESK_PROCESS_STACK_SIZE);
    CPPTESK_ASSERTION(ucp->uc_stack.ss_sp != NULL, "No memory");
}

Process::~Process()
{
    if(input_message != 0 && input_message != original_message)
        { delete input_message; }

    free(ucp->uc_stack.ss_sp);
    free(ucp);
}

bool greaterProcessPriority(Process *lhs, Process *rhs)
{
    priority_t lp = lhs->getPriority();
    priority_t rp = rhs->getPriority();

    // PIDs are used to sort equivalent processes in order of increasing PIDs.
    return lp > rp || (lp == rp && lhs->getPID() < rhs->getPID());
}

bool lessProcessPriority(Process *lhs, Process *rhs)
{
    priority_t lp = lhs->getPriority();
    priority_t rp = rhs->getPriority();

    return lp < rp || (lp == rp && lhs->getPID() < rhs->getPID());
}

}} // namespace cpptesk::hw

